!> @file user_actions.f90
!------------------------------------------------------------------------------!
! This file is part of PALM.
!
! PALM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2017 Leibniz Universitaet Hannover
!------------------------------------------------------------------------------!
!
! Current revisions:
! ------------------
! 
! 
! Former revisions:
! -----------------
! $Id: user_actions.f90 2101 2017-01-05 16:42:31Z suehring $
!
! 2000 2016-08-20 18:09:15Z knoop
! Forced header and separation lines into 80 columns
! 
! 1960 2016-07-12 16:34:24Z suehring
! New CASE statement for scalar tendency
! 
! 1873 2016-04-18 14:50:06Z maronga
! Module renamed (removed _mod)
! 
! 
! 1850 2016-04-08 13:29:27Z maronga
! Module renamed
!
! 
! 1822 2016-04-07 07:49:42Z hoffmann
! qr/nr-tendency removed.
!
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1353 2014-04-08 15:21:23Z heinze
! REAL constants provided with KIND-attribute
! 
! 1320 2014-03-20 08:40:49Z raasch
! kind-parameters added to all INTEGER and REAL declaration statements,
! kinds are defined in new module kinds,
! revision history before 2012 removed,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements 
!
! 1318 2014-03-17 13:35:16Z raasch
! module interfaces removed
!
! 1053 2012-11-13 17:11:03Z hoffmann
! +qr-tendency, nr-tendency
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 211 2008-11-11 04:46:24Z raasch
! Former file user_interface.f90 split into one file per subroutine
!
! Description:
! ------------
!> Execution of user-defined actions before or after single timesteps
!------------------------------------------------------------------------------!
 MODULE user_actions_mod
 

    PRIVATE
    PUBLIC user_actions

    INTERFACE user_actions
       MODULE PROCEDURE user_actions
       MODULE PROCEDURE user_actions_ij
    END INTERFACE user_actions

 CONTAINS


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Call for all grid points
!------------------------------------------------------------------------------!
    SUBROUTINE user_actions( location )

       USE constants

       USE control_parameters

       USE cpulog

       USE grid_variables

       USE indices

       USE kinds

       USE pegrid

       USE user

       USE arrays_3d

       USE surface_mod

       IMPLICIT NONE

       CHARACTER (LEN=*) ::  location !< 

       INTEGER(iwp) ::  i !< 
       INTEGER(iwp) ::  j !< 
       INTEGER(iwp) ::  k !< 
       INTEGER(iwp) ::  m !< 

       INTEGER(iwp) ::  ibit0, ibit1, ibit2, ibit3, ibit4, ibit5, ibit6, ibit7,&
                        ibit8
       INTEGER(iwp) ::  k_pp, k_ppp, k_mm

       REAL(wp)     ::  mask_ji, mask_east, mask_north, mask_top, ddt_3d

       CALL cpu_log( log_point(24), 'user_actions', 'start' )

!
!--    Here the user-defined actions follow
!--    No calls for single grid points are allowed at locations before and
!--    after the timestep, since these calls are not within an i,j-loop
       SELECT CASE ( location )

          CASE ( 'before_timestep' )
!
!--          Enter actions to be done before every timestep here
             s_previous = s

          CASE ( 'after_integration' )
!
!--          Calculate resolved-scalar scalar fluxes according to discretization
!--          in advection scheme
             DO  i = nxl, nxr
                DO  j = nys, nyn
                   wsss_res(nzb,j,i) = 0.0_wp
                   DO  k = nzb+1, nzt
!
!--                   Compute u's'
                      ibit2 = IBITS(advc_flags_1(k,j,i),2,1)
                      ibit1 = IBITS(advc_flags_1(k,j,i),1,1)
                      ibit0 = IBITS(advc_flags_1(k,j,i),0,1)

                      usss_res(k,j,i) = ( u(k,j,i+1) - u_gtrans ) *            &
                               (                                               &
                     ( 37.0_wp * ibit2 * adv_sca_5                             &
                  +     7.0_wp * ibit1 * adv_sca_3                             &
                  +              ibit0 * adv_sca_1                             &
                     ) *                                                       &
                             ( s(k,j,i+1) + s(k,j,i)   )                       &
              -      (  8.0_wp * ibit2 * adv_sca_5                             &
                  +              ibit1 * adv_sca_3                             &
                     ) *                                                       &
                             ( s(k,j,i+2) + s(k,j,i-1) )                       &
              +      (           ibit2 * adv_sca_5                             &
                     ) *                                                       &
                             ( s(k,j,i+3) + s(k,j,i-2) )                       &
                               )

                      usss_res(k,j,i) = usss_res(k,j,i) - ABS( u(k,j,i+1) - u_gtrans ) &
                                     * (                                       &
                     ( 10.0_wp * ibit2 * adv_sca_5                             &
                  +     3.0_wp * ibit1 * adv_sca_3                             &
                  +              ibit0 * adv_sca_1                             &
                     ) *                                                       &
                             ( s(k,j,i+1) - s(k,j,i)  )                        &
              -      (  5.0_wp * ibit2 * adv_sca_5                             &
                  +              ibit1 * adv_sca_3                             &
                     ) *                                                       &
                             ( s(k,j,i+2) - s(k,j,i-1) )                       &
              +      (           ibit2 * adv_sca_5                             &
                     ) *                                                       &
                             ( s(k,j,i+3) - s(k,j,i-2) )                       &
                                       )
!
!--                   Compute v's'
                      ibit5 = IBITS(advc_flags_1(k,j,i),5,1)
                      ibit4 = IBITS(advc_flags_1(k,j,i),4,1)
                      ibit3 = IBITS(advc_flags_1(k,j,i),3,1)

                      vsss_res(k,j,i) = v(k,j+1,i) - v_gtrans *                &
                               (                                               &
                     ( 37.0_wp * ibit5 * adv_sca_5                             &
                  +     7.0_wp * ibit4 * adv_sca_3                             &
                  +              ibit3 * adv_sca_1                             &
                     ) *                                                       &
                             ( s(k,j+1,i) + s(k,j,i)   )                       &
              -      (  8.0_wp * ibit5 * adv_sca_5                             &
                  +              ibit4 * adv_sca_3                             &
                     ) *                                                       &
                             ( s(k,j+2,i) + s(k,j-1,i) )                       &
              +      (           ibit5 * adv_sca_5                             &
                     ) *                                                       &
                             ( s(k,j+3,i) + s(k,j-2,i) )                       &
                               )

                      vsss_res(k,j,i) = vsss_res(k,j,i) - ABS( v(k,j+1,i) - v_gtrans ) &
                                     * (                                       &
                     ( 10.0_wp * ibit5 * adv_sca_5                             &
                  +     3.0_wp * ibit4 * adv_sca_3                             &
                  +              ibit3 * adv_sca_1                             &
                     ) *                                                       &
                             ( s(k,j+1,i) - s(k,j,i)   )                       &
              -      (  5.0_wp * ibit5 * adv_sca_5                             &
                  +              ibit4 * adv_sca_3                             &
                     ) *                                                       &
                             ( s(k,j+2,i) - s(k,j-1,i) )                       &
              +      (           ibit5 * adv_sca_5                             &
                     ) *                                                       &
                             ( s(k,j+3,i) - s(k,j-2,i) )                       &
                                       )

!
!--                   Compute w's'
                      ibit8 = IBITS(advc_flags_1(k,j,i),8,1)
                      ibit7 = IBITS(advc_flags_1(k,j,i),7,1)
                      ibit6 = IBITS(advc_flags_1(k,j,i),6,1)

                      k_ppp = k + 3 * ibit8
                      k_pp  = k + 2 * ( 1 - ibit6  )
                      k_mm  = k - 2 * ibit8

                      wsss_res(k,j,i) = w(k,j,i) * (                           &
                     ( 37.0_wp * ibit8 * adv_sca_5                             &
                  +     7.0_wp * ibit7 * adv_sca_3                             &
                  +              ibit6 * adv_sca_1                             &
                     ) *                                                       &
                             ( s(k+1,j,i)  + s(k,j,i)    )                     &
              -      (  8.0_wp * ibit8 * adv_sca_5                             &
                  +              ibit7 * adv_sca_3                             &
                     ) *                                                       &
                             ( s(k_pp,j,i) + s(k-1,j,i)  )                     &
              +      (           ibit8 * adv_sca_5                             &
                     ) *     ( s(k_ppp,j,i)+ s(k_mm,j,i) )                     &
                                 )

                      wsss_res(k,j,i) = wsss_res(k,j,i) - ABS( w(k,j,i) ) * (  &
                     ( 10.0_wp * ibit8 * adv_sca_5                             &
                  +     3.0_wp * ibit7 * adv_sca_3                             &
                  +              ibit6 * adv_sca_1                             &
                     ) *                                                       &
                             ( s(k+1,j,i)   - s(k,j,i)    )                    &
              -      (  5.0_wp * ibit8 * adv_sca_5                             &
                  +              ibit7 * adv_sca_3                             &
                     ) *                                                       &
                             ( s(k_pp,j,i)  - s(k-1,j,i)  )                    &
              +      (           ibit8 * adv_sca_5                             &
                     ) *                                                       &
                             ( s(k_ppp,j,i) - s(k_mm,j,i) )                    &
                                                                    )
                                                                    
                      div_res(k,j,i) = (                                      &
                          ( u(k,j,i+1) - u(k,j,i)   ) * ddx                   &
                        + ( v(k,j+1,i) - v(k,j,i)   ) * ddy                   &
                        + ( w(k,j,i)   - w(k-1,j,i) ) * ddzw(k)               &
                                       ) * s(k,j,i)

!
!--                   Predetermine flag to mask wall-bounded grid points
                      mask_ji    = MERGE( 1.0_wp, 0.0_wp, BTEST( wall_flags_0(k,j,i), 0   ) )
                      mask_east  = MERGE( 1.0_wp, 0.0_wp, BTEST( wall_flags_0(k,j,i+1), 0 ) )
                      mask_north = MERGE( 1.0_wp, 0.0_wp, BTEST( wall_flags_0(k,j+1,i), 0 ) )
                      mask_top   = MERGE( 1.0_wp, 0.0_wp, BTEST( wall_flags_0(k+1,j,i), 8 ) )

                      usss_sgs(k,j,i) = 0.5_wp * mask_ji *                     &
                            mask_east  * ( kh(k,j,i) + kh(k,j,i+1) )           &
                                       * ( s(k,j,i+1) - s(k,j,i)   ) * ddx

                      vsss_sgs(k,j,i) = 0.5_wp * mask_ji *                     &
                            mask_north * ( kh(k,j,i) + kh(k,j+1,i) )           &
                                       * ( s(k,j+1,i) - s(k,j,i)   ) * ddy

                      wsss_sgs(k,j,i) = 0.5_wp * mask_ji *                     &
                                      ( kh(k,j,i) + kh(k+1,j,i) ) *            &
                                          ( s(k+1,j,i)-s(k,j,i) ) * ddzu(k+1)  &
                                                            * mask_top
!
!--                   Horizontal wind velocity
                      ff_h(k,j,i) = SQRT(                                      &
                                  ( ( u(k,j,i) + u(k,j,i+1) ) * 0.5 )**2 +     &
                                  ( ( v(k,j,i) + v(k,j+1,i) ) * 0.5 )**2       &
                                      )
!
!--                   3D wind velocity
                      ff(k,j,i) = SQRT(                                        &
                                  ( ( u(k,j,i) + u(k,j,i+1) ) * 0.5 )**2 +     &
                                  ( ( v(k,j,i) + v(k,j+1,i) ) * 0.5 )**2 +     &
                                  ( ( w(k,j,i) + v(k-1,j,i) ) * 0.5 )**2       &
                                      )
!
!--                   u*u
                      uu(k,j,i) = u(k,j,i) * u(k,j,i)
!
!--                   v*v
                      vv(k,j,i) = v(k,j,i) * v(k,j,i)
!
!--                   w*w
                      ww(k,j,i) = w(k,j,i) * w(k,j,i)
!
!--                   left flow
                      u_plus(k,j,i) = 0.5_wp * ( u(k,j,i) + ABS( u(k,j,i) ) )
!
!--                   right flow
                      u_minus(k,j,i) = 0.5_wp * ( u(k,j,i) - ABS( u(k,j,i) ) )
!
!--                   northern flow
                      v_plus(k,j,i) = 0.5_wp * ( v(k,j,i) + ABS( v(k,j,i) ) )
!
!--                   southern flow
                      v_minus(k,j,i) = 0.5_wp * ( v(k,j,i) - ABS( v(k,j,i) ) )
!
!--                   upward flow
                      w_plus(k,j,i) = 0.5_wp * ( w(k,j,i) + ABS( w(k,j,i) ) )
!
!--                   downward flow
                      w_minus(k,j,i) = 0.5_wp * ( w(k,j,i) - ABS( w(k,j,i) ) )
!
!--                   turbulence intensity: rot[(u,v,w)]
                      ti(k,j,i) = 0.25_wp * SQRT(                           &
                        (   (   w(k,j+1,i) + w(k-1,j+1,i)                   &
                              - w(k,j-1,i) - w(k-1,j-1,i) ) * ddy           &
                          - (   v(k+1,j,i) + v(k+1,j+1,i)                   &
                              - v(k-1,j,i) - v(k-1,j+1,i) ) * ddzu(k) )**2  &
                      + (   (   u(k+1,j,i) + u(k+1,j,i+1)                   &
                              - u(k-1,j,i) - u(k-1,j,i+1) ) * ddzu(k)       &
                          - (   w(k,j,i+1) + w(k-1,j,i+1)                   &
                              - w(k,j,i-1) - w(k-1,j,i-1) ) * ddx     )**2  &
                      + (   (   v(k,j,i+1) + v(k,j+1,i+1)                   &
                              - v(k,j,i-1) - v(k,j+1,i-1) ) * ddx           &
                          - (   u(k,j+1,i) + u(k,j+1,i+1)                   &
                              - u(k,j-1,i) - u(k,j-1,i+1) ) * ddy     )**2  )
                   ENDDO
                ENDDO
             ENDDO
!
!--          For the subgrid-scale parts, add the respective fluxes at the surfaces
!
!--          Default-type surfaces, upward-facing
             DO  m = 1, surf_def_h(0)%ns
                i = surf_def_h(0)%i(m)
                j = surf_def_h(0)%j(m)
                k = surf_def_h(0)%k(m)

                wsss_sgs(k-1,j,i) = surf_def_h(0)%ssws(m)
             ENDDO
!
!--          Default-type surfaces, downward-facing
             DO  m = 1, surf_def_h(1)%ns
                i = surf_def_h(1)%i(m)
                j = surf_def_h(1)%j(m)
                k = surf_def_h(1)%k(m)

                wsss_sgs(k+1,j,i) = surf_def_h(1)%ssws(m)
             ENDDO
!
!--          Compute balance terms. Before doing this, exchange ghost points.
             CALL exchange_horiz( usss_res, nbgp )
             CALL exchange_horiz( vsss_res, nbgp )
             CALL exchange_horiz( wsss_res, nbgp )
             CALL exchange_horiz( usss_sgs, nbgp )
             CALL exchange_horiz( vsss_sgs, nbgp )
             CALL exchange_horiz( wsss_sgs, nbgp )
             CALL exchange_horiz( div_res,  nbgp )
             

             ddt_3d = 1.0_wp / dt_3d
             DO  i = nxl, nxr
                DO  j = nys, nyn
                   DO  k = nzb+1, nzt
                      d_usss_res(k,j,i) = ( usss_res(k,j,i) - usss_res(k,j,i-1) ) * ddx
                      d_vsss_res(k,j,i) = ( vsss_res(k,j,i) - vsss_res(k,j-1,i) ) * ddy 
                      d_wsss_res(k,j,i) = ( wsss_res(k,j,i) - wsss_res(k-1,j,i) ) * ddzw(k)

                      d_usss_sgs(k,j,i) = ( usss_sgs(k,j,i) - usss_sgs(k,j,i-1) ) * ddx
                      d_vsss_sgs(k,j,i) = ( vsss_sgs(k,j,i) - vsss_sgs(k,j-1,i) ) * ddy 
                      d_wsss_sgs(k,j,i) = ( wsss_sgs(k,j,i) - wsss_sgs(k-1,j,i) ) * ddzw(k)

                      dsdt(k,j,i) = ( s(k,j,i) - s_previous(k,j,i) ) * ddt_3d
                   ENDDO
                ENDDO
             ENDDO
!
!--          Note, no wallfluxes at vertical walls for scalar given!



          CASE ( 'after_timestep' )
!
!--          Enter actions to be done after every timestep here


          CASE ( 'u-tendency' )
!
!--          Enter actions to be done in the u-tendency term here


          CASE ( 'v-tendency' )


          CASE ( 'w-tendency' )


          CASE ( 'pt-tendency' )


          CASE ( 'sa-tendency' )


          CASE ( 'e-tendency' )


          CASE ( 'q-tendency' )
          
          
          CASE ( 's-tendency' )          


          CASE DEFAULT
             message_string = 'unknown location "' // location // '"'
             CALL message( 'user_actions', 'UI0001', 1, 2, 0, 6, 0 )

       END SELECT

       CALL cpu_log( log_point(24), 'user_actions', 'stop' )

    END SUBROUTINE user_actions


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Call for grid point i,j
!------------------------------------------------------------------------------!
    SUBROUTINE user_actions_ij( i, j, location )

       USE control_parameters
       USE kinds
       USE pegrid
       USE user

       IMPLICIT NONE

       CHARACTER (LEN=*) ::  location

       INTEGER(iwp) ::  i
       INTEGER(iwp) ::  idum
       INTEGER(iwp) ::  j

!
!--    Here the user-defined actions follow
       SELECT CASE ( location )

          CASE ( 'u-tendency' )
!
!--          Enter actions to be done in the u-tendency term here


          CASE ( 'v-tendency' )


          CASE ( 'w-tendency' )


          CASE ( 'pt-tendency' )


          CASE ( 'sa-tendency' )


          CASE ( 'e-tendency' )


          CASE ( 'q-tendency' )
          
          
          CASE ( 's-tendency' )


          CASE ( 'before_timestep', 'after_integration', 'after_timestep' )
             message_string = 'location "' // location // '" is not ' // &
                             'allowed to be called with parameters "i" and "j"'
             CALL message( 'user_actions', 'UI0002', 1, 2, 0, 6, 0 )


          CASE DEFAULT
             message_string = 'unknown location "' // location // '"'
             CALL message( 'user_actions', 'UI0001', 1, 2, 0, 6, 0 )
             

       END SELECT

    END SUBROUTINE user_actions_ij

 END MODULE user_actions_mod
